package phylogenetics

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Phylogenetics.MG94Source

Codon rate matrix from Muse and Gaut 94

This modules provides functions to build rates matrices for codon evolution following the model proposed by Muse and Gaut 1. The exact formulation follows the parameterization used by Latrille and Lartillot 2.

References:

Sourcemodule NSCodon_rate_matrix : sig ... end
Sourcetype param = {
  1. nucleotide_rates : Rate_matrix.Nucleotide.t;
  2. nucleotide_stat_dist : Nucleotide.vector;
  3. omega : float;
}
Sourceval rate_matrix : param -> NSCodon_rate_matrix.t
Sourceval stationary_distribution : param -> NSCodon.vector